import hashlib
import logging
from typing import Any, Dict, List, Optional

from solitude import TOOL_NAME
from solitude.slurmjob import SlurmJob
from solitude.cache import Cache
from solitude import hookimpl


def md5(s) -> str:
    return hashlib.md5(s.encode()).hexdigest()


class CommandBase(object):
    def __init__(
        self, cmd: str, plugin: hookimpl, cache: Optional[Cache] = None
    ):
        self.logger = logging.getLogger(type(self).__name__)
        self._plugin = plugin
        self.cmd = cmd.strip()
        self.cache = cache
        self.hash = md5(
            self.get_plugin_short_name()
            + self._plugin.get_command_hash(self.cmd)
        )
        self.job_info = None
        self._set_basic_job_info_from_cache()
        self._state: Dict[str, Any] = {}
        self.errors: List[str] = []

    def get_plugin_short_name(self) -> str:
        return self._plugin.__name__.rsplit(".", 2)[-1].replace(
            f"{TOOL_NAME}_", ""
        )

    def update(self):
        self.update_job_info()
        self.update_state()
        self.update_errors()

    def update_state(self):
        self._state = self._plugin.retrieve_state(self.cmd)

    def update_errors(self):
        self.errors = self._update_errors()

    def _set_basic_job_info_from_cache(self):
        self.job_info = None
        if self.cache is not None and self.hash in self.cache:
            entry = self.cache[self.hash]
            self.job_info = entry

    def update_job_info(self):
        if self.cache is not None and self.hash in self.cache:
            try:
                job_info = SlurmJob(jobid=self.cache[self.hash].id)
                job_info.update()
                self.job_info = job_info
            except ValueError:
                self.job_info = None
            except Exception as e:
                self.logger.warning(
                    "Job with id: {} has no job_info: {}".format(self.hash, e)
                )
                raise e

    def _update_errors(self):
        if (
            self.has_job_link()
            and not self.job_info.is_running()
            and not self.is_finished()
        ):
            try:
                joblog = self.job_info.get_log_text()
                errors = self._plugin.get_errors_from_log(log=joblog)
                return errors
            except Exception as e:
                self.logger.error(
                    f"Exception while fetching errors from log: {e}"
                )
        return []

    def has_job_link(self) -> bool:
        return self.job_info is not None

    def get_job_status_str(self) -> str:
        if self.job_info is not None:
            if self.job_info.is_pending():
                return "PEND"
            elif self.job_info.is_timeout():
                return "TIME"
            elif self.is_running():
                return "RUN"
        if self.is_erroneous():
            return "ERR!"
        elif self.is_finished():
            return "."
        return "IDLE"

    def __str__(self) -> str:
        self._errors: List[str] = []
        url, priority = "-", ""
        if self.job_info is not None:
            url = self.job_info.get_log_url()
            if self.is_running() and self.job_info.priority == "high":
                priority = "H*"
        return "{status:4}{priority:2} {url:28} {plugin:10} {tag} {errors}".format(
            plugin=self.get_plugin_short_name(),
            tag=self.get_job_info_str(),
            status=self.get_job_status_str(),
            url=url,
            priority=priority,
            errors=f'!!ERRORS: {" ".join(self.errors)}!!'
            if self.is_erroneous()
            else "",
        )

    @staticmethod
    def header_str() -> str:
        return (
            "id  {status:6} {url:28} {plugin:10} {info}".format(
                plugin="plugin",
                info="job_info",
                status="status",
                url="log_url",
            )
            + "\n"
            + "--  {status:6} {url:28} {plugin:10} {info}".format(
                plugin="------",
                info="--------",
                status="------",
                url="-------",
            )
        )

    def is_running(self) -> bool:
        return self.job_info is not None and self.job_info.is_running()

    def is_finished(self) -> bool:
        return self._plugin.is_command_job_done(
            cmd=self.cmd, state=self._state
        )

    def is_erroneous(self) -> bool:
        return any(self.errors)

    def get_job_info_str(self) -> str:
        return self._plugin.get_command_status_str(
            cmd=self.cmd, state=self._state
        )



from typing import Dict, List
from solitude import hookimpl
from solitude.plugins.csubmit import (
    get_errors_from_log as _get_errors_from_log,
)
from solitude.plugins.csubmit import (
    filter_csubmit_command_str,
    match_csubmit_command_str,
)


@hookimpl
def matches_command(cmd: str) -> bool:
    return (" --interactive " in cmd) and (
        match_csubmit_command_str(cmd=cmd) is not None
    )


@hookimpl
def get_command_hash(cmd: str) -> str:
    return filter_csubmit_command_str(cmd=cmd)


@hookimpl
def retrieve_state(cmd: str) -> Dict:
    return {}


@hookimpl
def is_command_job_done(cmd: str, state: Dict) -> bool:
    return False


@hookimpl
def get_command_status_str(cmd: str, state: Dict) -> str:
    return filter_csubmit_command_str(cmd=cmd)


@hookimpl
def get_errors_from_log(log: str) -> List[str]:
    return _get_errors_from_log(log=log)



import pluggy

__author__ = "Sil van de Leemput"
__email__ = "sil.vandeleemput@radboudumc.nl"
__version__ = "0.2.0"

TOOL_AUTHOR = "DIAG"
TOOL_NAME = "solitude"
hookimpl = pluggy.HookimplMarker(TOOL_NAME)



import re
from typing import Dict, List, Optional
from solitude import hookimpl


def match_csubmit_command_str(cmd: str) -> Optional[re.Match]:
    return re.search(
        r"c-submit\s+(?:(?:--\S+(?:\s*[=\s]\s*\S+)?\s+)*)?(?P<user>\S+)"
        r"\s+(?P<ticket>\S+)\s+(?P<duration>\d+)\s+(?P<image>\D\S+)\s*(?P<args>.*)",
        cmd.strip(),
    )


def filter_csubmit_command_str(cmd: str) -> str:
    result = match_csubmit_command_str(cmd=cmd)
    assert result is not None
    return (
        result.groupdict()["image"] + " " + result.groupdict()["args"]
    ).strip()


@hookimpl
def matches_command(cmd: str) -> bool:
    return match_csubmit_command_str(cmd=cmd) is not None


@hookimpl
def get_command_hash(cmd: str) -> str:
    return filter_csubmit_command_str(cmd=cmd)


@hookimpl
def retrieve_state(cmd: str) -> Dict:
    return {}


@hookimpl
def is_command_job_done(cmd: str, state: Dict) -> bool:
    return False


@hookimpl
def get_command_status_str(cmd: str, state: Dict) -> str:
    return filter_csubmit_command_str(cmd=cmd)


@hookimpl
def get_errors_from_log(log: str) -> List[str]:
    errors = []
    if log.find("Killed") != -1:
        errors.append("NoResource")
    if log.find("CUDA_ERROR") != -1:
        errors.append("CUDA")
    errors = errors + [e[1:-1] for e in re.findall(r"\s.*Error:", log)]
    return errors



import collections
import copy
import getpass
import json
import logging
import pprint
import re
from pathlib import Path
from typing import Any, Dict, List

import click

import solitude.core
from solitude.core import log_setup
from solitude.config import get_config
from solitude.cache import json_write_atomic
from solitude import TOOL_NAME, __version__
from solitude.slurmjob import SlurmJob
from solitude.utils.ssh import SSHClient


DEFAULT_SSH_SERVER = "dlc-umbriel.umcn.nl"


def resolve_jobids(
    ctx: click.core.Context, param: click.core.Option, value: Any
) -> List[int]:
    # split values out on " " and ","
    substrs = [
        v3
        for v in value
        for v2 in v.strip().split(" ")
        for v3 in v2.strip().split(",")
    ]
    indices = []
    for entry in substrs:
        res = re.match(r"(\d+)(?=-(\d+))?", entry)
        if not res:
            click.echo(
                "Indices should be specified using numbers, "
                "ranges (a-b) or as a list of comma separated indices"
            )
            ctx.exit()
        grps = res.groups()
        if grps[1] is None:
            # parse number
            indices.append(int(grps[0]))
        else:
            # parse range
            a, b = int(grps[0]), int(grps[1])
            if a > b:
                click.echo(
                    "The second value (b) in a range should "
                    "be greater or equal to the first value (a) (b >= a)"
                )
                ctx.exit()
            indices = indices + [e for e in range(a, b + 1)]
    indices = [e for e in sorted(set(indices))]
    if not all(map(lambda x: x > 0, indices)):
        click.echo("Jobids should be greater than 0")
        ctx.exit()
    return indices


def is_config_valid() -> bool:
    cfg = get_config()
    if not cfg.is_config_present():
        click.echo(
            "Configuration file could not be found. "
            "Generate one using `solitude config create`"
        )
        return False
    if not cfg.is_ssh_configured():
        click.echo(
            f"Configuration file was found at {cfg.config_path}. "
            f"However ssh was not properly configured. "
            "Generate new settings using `solitude config create`"
        )
        return False
    return True


def print_version(
    ctx: click.core.Context, param: click.core.Option, value: Any
):
    if not value or ctx.resilient_parsing:
        return
    click.echo(f"{TOOL_NAME} {__version__}")
    ctx.exit()


def set_verbose_logging(
    ctx: click.core.Context, param: click.core.Option, value: Any
):
    if not value or ctx.resilient_parsing:
        return
    logger = logging.getLogger(f"{TOOL_NAME}")
    logger.setLevel(logging.DEBUG)


class ClickPath(click.Path):
    """A Click path argument that returns a pathlib Path, not a string"""

    def convert(self, value, param, ctx):
        return Path(super().convert(value, param, ctx))


class BaseGroup(click.core.Group):
    def __init__(self, name=None, commands=None, **kwargs):
        commands = commands or collections.OrderedDict()
        super().__init__(name, commands, **kwargs)
        self.params.insert(
            200,
            click.core.Option(
                param_decls=("--version",),
                is_flag=True,
                is_eager=True,
                expose_value=False,
                callback=print_version,
                help="Show the version and exit.",
            ),
        )

    def list_commands(
        self, ctx: click.core.Context
    ) -> Dict[str, click.core.Command]:
        return self.commands


class BaseCommand(click.core.Command):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.params.insert(
            199,
            click.core.Option(
                param_decls=("-v", "--verbose"),
                is_flag=True,
                is_eager=True,
                expose_value=False,
                callback=set_verbose_logging,
                help="Enable verbose messages.",
            ),
        )
        self.params.insert(
            200,
            click.core.Option(
                param_decls=("--version",),
                is_flag=True,
                is_eager=True,
                expose_value=False,
                callback=print_version,
                help="Show the version and exit.",
            ),
        )


class SharedCommand(BaseCommand):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.params.insert(
            0,
            click.core.Option(
                param_decls=("-f", "--cmd-files"),
                multiple=True,
                required=True,
                type=ClickPath(dir_okay=True, exists=True),
                default=get_config().defaults.cmdfiles,
                help=f"File(s) containing a list of commands. Default: {get_config().defaults.cmdfiles}",
            ),
        )
        self.params.insert(
            1,
            click.core.Option(
                param_decls=("-i", "--jobids"),
                multiple=True,
                type=str,
                callback=resolve_jobids,
                help="Select specific jobids. E.g.: 1-5,7,9",
            ),
        )
        self.params.insert(
            2,
            click.core.Option(
                param_decls=("-w", "--workers"),
                default=get_config().defaults.workers,
                help=f"Workers to use for scheduling. Default: {get_config().defaults.workers}",
            ),
        )


@click.group(
    cls=BaseGroup,
    help="A tool for running and managing commands on the DIAG SOL cluster",
)
def cli():
    pass


def are_jobids_specified(jobids: List[int]) -> bool:
    are_specified = len(jobids) > 0
    if not are_specified:
        click.echo(
            "This operation requires specification of jobids with -i or --jobids"
        )
    return are_specified


@click.command(cls=SharedCommand, help="Run jobs")
@click.option(
    "-u",
    "--user",
    default=get_config().defaults.username,
    help="Run command job(s) for the given user.",
)
@click.option(
    "-p",
    "--priority",
    default=get_config().defaults.priority,
    type=click.Choice(["idle", "low", "high"], case_sensitive=False),
    help="Run command job(s) with priority level (idle|low|high)",
)
@click.option(
    "-r",
    "--reservation",
    default=None,
    help="Run command job(s) for the given reservation.",
)
@click.option(
    "-d",
    "--duration",
    default=None,
    type=int,
    help="Run command job(s) for the given duration.",
)
@click.option(
    "-x",
    "--ignore-errors",
    default=False,
    flag_value=True,
    help="Run command job(s) despite reported errors.",
)
def run(
    cmd_files: List[Path],
    jobids: List[int],
    workers: int,
    user: str,
    priority: str,
    reservation: str,
    duration: int,
    ignore_errors: bool,
):
    if are_jobids_specified(jobids=jobids) and is_config_valid():
        solitude.core.run(
            cmd_files=cmd_files,
            jobids=jobids,
            workers=workers,
            user=user,
            priority=priority,
            reservation=reservation,
            duration=duration,
            ignore_errors=ignore_errors,
            cfg=get_config(),
            cache_file=get_config().cache_path,
        )


@click.command(cls=SharedCommand, help="Extend running jobs")
def extend(cmd_files: List[Path], jobids: List[int], workers: int):
    if are_jobids_specified(jobids=jobids) and is_config_valid():
        solitude.core.extend(
            cmd_files=cmd_files,
            jobids=jobids,
            workers=workers,
            cfg=get_config(),
            cache_file=get_config().cache_path,
        )


@click.command(cls=SharedCommand, help="Stop running jobs")
def stop(cmd_files: List[Path], jobids: List[int], workers: int):
    if are_jobids_specified(jobids=jobids) and is_config_valid():
        solitude.core.stop(
            cmd_files=cmd_files,
            jobids=jobids,
            workers=workers,
            cfg=get_config(),
            cache_file=get_config().cache_path,
        )


@click.command(cls=SharedCommand, help="Link jobs to existing slurm job")
@click.option(
    "-s",
    "--slurm-jobid",
    "slurm_jobid",
    required=True,
    type=int,
    help="Slurm job id to link the jobs to.",
)
def link(
    cmd_files: List[Path], jobids: List[int], workers: int, slurm_jobid: int
):
    if are_jobids_specified(jobids=jobids) and is_config_valid():
        if SlurmJob.check_if_job_exists(jobid=slurm_jobid):
            for jobid in jobids:
                success = solitude.core.link(
                    cmd_files=cmd_files,
                    jobid=jobid,
                    slurm_jobid=slurm_jobid,
                    cfg=get_config(),
                    cache_file=get_config().cache_path,
                )
                click.echo(
                    f"Linked job id {jobid} to slurm job: {slurm_jobid} {'succeeded' if success else 'failed'}"
                )
        else:
            click.echo(f"Slurm job id: {slurm_jobid} does not exists...")


@click.command(cls=SharedCommand, help="Unlink jobs from slurm job")
def unlink(cmd_files: List[Path], jobids: List[int], workers: int):
    if are_jobids_specified(jobids=jobids) and is_config_valid():
        for jobid in jobids:
            success = solitude.core.unlink(
                cmd_files=cmd_files,
                jobid=jobid,
                cfg=get_config(),
                cache_file=get_config().cache_path,
            )
            if success:
                click.echo(f"Unlinked job id {jobid} from slurm jobs.")
            else:
                click.echo(
                    f"Job id {jobid} has no slurm job associated with it."
                )


@click.command(name="list", cls=SharedCommand, help="List jobs")
@click.option(
    "-s",
    "--selected-only",
    "selected_only",
    flag_value=True,
    default=False,
    help="Only display the jobs indicated by jobids. Default=No",
)
@click.option(
    "-np",
    "--no-progress",
    "show_progress_bar",
    flag_value=False,
    default=True,
    help="Hide progress bar while loading the list. Default=No",
)
def list_jobs(
    cmd_files: List[Path],
    jobids: List[int],
    workers: int,
    selected_only: bool,
    show_progress_bar: bool,
):
    if is_config_valid():
        solitude.core.list_jobs(
            cmd_files=cmd_files,
            jobids=jobids,
            workers=workers,
            selected_only=selected_only,
            show_progress_bar=show_progress_bar,
            cfg=get_config(),
            cache_file=get_config().cache_path,
        )


@click.group(cls=BaseGroup, help="Job operations")
def job():
    pass


@click.group(cls=BaseGroup, help="Configuration operations")
def config():
    pass


@click.command(cls=BaseCommand, help="Show status of the configuration")
def status():
    cfg = get_config()
    is_config_valid()
    if cfg.is_config_present():
        click.echo(f"Configuration file found at: {cfg.config_path}")
        with open(cfg.config_path, "r") as f:
            cfg = json.load(f)
        display_configuration(cfg=cfg)


@click.command(cls=BaseCommand, help="Create or refresh the configuration")
@click.option(
    "--user",
    prompt="Default username for running jobs using SOL",
    type=str,
    required=True,
    default=getpass.getuser(),
    help="Set default user to run jobs on SOL.",
)
@click.option(
    "--workers",
    type=int,
    default=8,
    help="Set default number of workers for issuing jobs on the server.",
)
@click.option(
    "--priority",
    type=click.Choice(["idle", "low", "high"]),
    default="low",
    help="Set default priority to run jobs with.",
)
@click.option(
    "--cmd-files",
    type=ClickPath(exists=True, dir_okay=False),
    multiple=True,
    default=(),
    help="Set default command file(s) to use.",
)
@click.option(
    "--ssh-server",
    prompt="Default ssh DIAG deep learning server to use",
    type=str,
    default=DEFAULT_SSH_SERVER,
    required=True,
    help="Set default ssh DIAG deep learning server. Any of: `dlc-{MACHINE}.umcn.nl`.",
)
@click.option(
    "--ssh-user",
    prompt="Default ssh username to use",
    type=str,
    default="diag",
    required=True,
    help="Set default ssh username.",
)
@click.option(
    "--ssh-pass",
    prompt="Default ssh password to use",
    type=str,
    required=True,
    hide_input=True,
    confirmation_prompt=True,
    help="Set default ssh password.",
)
@click.option(
    "-i",
    "--ignore-test",
    default=False,
    flag_value=True,
    help="Ignore ssh connectivity test.",
)
@click.option(
    "-f",
    "--force",
    default=False,
    flag_value=True,
    help="Overwrite previous configuration file if it exists.",
)
def create(
    user: str,
    workers: int,
    priority: str,
    cmd_files: List[Path],
    ssh_server: str,
    ssh_user: str,
    ssh_pass: str,
    ignore_test: bool,
    force: bool,
):
    ssh_creds = {
        "server": ssh_server,
        "username": ssh_user,
        "password": ssh_pass,
    }
    cfgnew = {
        "defaults": {
            "username": user,
            "workers": workers,
            "priority": priority,
            "cmdfiles": list(map(str, cmd_files)),
        },
        "ssh": ssh_creds,
        "plugins": [],
    }
    if not ignore_test:
        connected = test_ssh_credentials(**ssh_creds)
        if not connected and not click.confirm(
            f"SSH connection failed using provided credentials. Do you want to keep these settings anyway?",
            default=False,
        ):
            return
    else:
        click.echo(f"TEST: CONNECT to server ({ssh_server}): SKIPPED...")

    cfg = get_config()
    if (
        not cfg.is_config_present()
        or force
        or click.confirm(
            f"There already is a configuration file at: {cfg.config_path}"
            f", Do you want to overwrite this file?",
            default=False,
        )
    ):
        click.echo(f"Writing configuration to: {cfg.config_path}")
        cfg.config_path.parent.mkdir(parents=True, exist_ok=True)
        json_write_atomic(fname=cfg.config_path, data=cfgnew)
        display_configuration(cfg=cfgnew)


@click.command(
    name="test",
    cls=BaseCommand,
    help="Test the configuration's ssh credentials",
)
def test_config():
    cfg = get_config()
    if not cfg.is_config_present():
        click.echo(
            "There is no configuration file present. Create one with: solitude config create"
        )
    else:
        test_ssh_credentials(*cfg.ssh)


def test_ssh_credentials(server: str, username: str, password: str) -> bool:
    connected = SSHClient(exit_on_fail=False).connect(
        server, username, password
    )
    click.echo(
        f"TEST: CONNECT to server ({server}): {'SUCCESS' if connected else 'FAILED!'}"
    )
    if not connected:
        click.echo(
            f"Are you connected to the SOL network (VPN)?\n"
            f"Are the provided ssh credentials correct?\n"
        )
    return connected


def display_configuration(cfg: Dict):
    """Display configuration, but obscure the ssh password"""
    cfgcopy = copy.deepcopy(cfg)
    if "ssh" in cfgcopy and "password" in cfgcopy["ssh"]:
        cfgcopy["ssh"]["password"] = "********"
    click.echo("Configuration dump:")
    pprint.pprint(cfgcopy)


job.add_command(list_jobs)
job.add_command(run)
job.add_command(extend)
job.add_command(stop)
job.add_command(link)
job.add_command(unlink)
cli.add_command(job)

config.add_command(status)
config.add_command(create)
config.add_command(test_config)
cli.add_command(config)


def main():
    logger = log_setup()
    logger.setLevel(logging.INFO)
    cli()


if __name__ == "__main__":
    main()



from typing import Dict, Optional, Type

import logging
import requests
import time

from solitude.cache import CacheData


DEFAULT_HOST = "http://oni:11080/"
logger = logging.getLogger("SlurmJob")


class SlurmJob(CacheData):
    def __init__(
        self,
        jobid: int,
        user: Optional[str] = None,
        priority: Optional[str] = None,
        timestamp: Optional[int] = None,
    ):
        super(SlurmJob, self).__init__(timestamp=timestamp)
        self.id: int = jobid
        self.user: Optional[str] = user
        self.priority: Optional[str] = priority
        self._status: Optional[str] = None

    def update(self):
        info = requests.get(f"{DEFAULT_HOST}json/jobinfo/{self.id}").json()
        self.id = info["jobid"]
        self.user = info["user"]
        self.priority = info["priority"]
        self._status = info["status"]

    def is_running(self) -> bool:
        return self._status in ("RUNNING", "PENDING")

    def is_pending(self) -> bool:
        return self._status == "PENDING"

    def is_timeout(self) -> bool:
        return self._status == "TIMEOUT"

    def get_log_text(self) -> str:
        return requests.get(
            f"{DEFAULT_HOST}text/log/{self.id}", allow_redirects=True
        ).text

    def get_log_url(self) -> str:
        return "{}show/{}".format(DEFAULT_HOST, self.id)

    def to_dict(self) -> Dict:
        return dict(
            id=self.id,
            user=self.user,
            priority=self.priority,
            timestamp=self.timestamp,
        )

    @staticmethod
    def check_if_job_exists(jobid: int) -> bool:
        txt = requests.get(f"{DEFAULT_HOST}json/jobinfo/{jobid}").text
        return "does not exist" not in txt

    @classmethod
    def from_dict(cls: Type, dic: Dict) -> "SlurmJob":
        return SlurmJob(
            jobid=dic["id"],
            user=dic["user"],
            priority=dic["priority"],
            timestamp=dic.get("timestamp", int(time.time())),
        )



import logging
import json
from enum import Enum, unique
from typing import Callable, Dict, List, NamedTuple, Optional, Tuple
from pathlib import Path

import appdirs

from solitude import TOOL_AUTHOR, TOOL_NAME

__config = None


def get_existing_filepath(
    filename: str, valid_dirs: List[Path], allow_non_existing: bool = False
) -> Path:
    assert any(valid_dirs)
    full_path = Path()
    for valid_dir in valid_dirs:
        full_path = valid_dir / filename
        if full_path.is_file():
            return full_path
    if allow_non_existing:
        return full_path
    raise FileNotFoundError(
        f"Could not find a `{filename}` file in: {valid_dirs}"
    )


def resolve_core_filepaths(
    allow_non_existing: bool = False,
) -> Tuple[Path, Optional[Path], Path]:
    # Look for configuration and cache file:
    # 1) locally
    # 2) User home dir `user_config_dir` or `user_cache_dir`
    local_dir = Path(__file__).parent.parent.absolute()
    dirs = appdirs.AppDirs(appname=TOOL_NAME, appauthor=TOOL_AUTHOR)
    cache_fname = get_existing_filepath(
        filename="cache.json",
        valid_dirs=[local_dir, Path(dirs.user_cache_dir)],
        allow_non_existing=True,
    )
    assert cache_fname is not None
    config_filepath = get_existing_filepath(
        filename="config.json",
        valid_dirs=[local_dir, Path(dirs.user_config_dir)],
        allow_non_existing=allow_non_existing,
    )
    return cache_fname, config_filepath, local_dir


class Config(object):
    class Defaults(NamedTuple):
        username: str = "UNKNOWN"
        priority: str = "low"
        workers: int = 8
        cmdfiles: Tuple = ()

    class SSHConfig(NamedTuple):
        server: str
        username: str
        password: str

    @unique
    class JobPriority(Enum):
        IDLE = "idle"
        LOW = "low"
        HIGH = "high"

    def __init__(self):
        self.logger = logging.getLogger(type(self).__name__)
        self.plugins = []
        self.defaults = Config.Defaults()
        self.ssh: Optional[
            Config.SSHConfig
        ] = None  # this is None if no valid ssh configuration is set
        self.config_path = None
        self.cache_path, self.config_path, _ = resolve_core_filepaths(
            allow_non_existing=True
        )
        self.load_config()

    def _validate_property(
        self,
        settings: Dict,
        group: str,
        prop: str,
        test: Optional[Callable] = None,
        msg: str = "must be a string value.",
    ) -> bool:
        passed: bool
        try:
            passed = (
                test(settings[group][prop])
                if test is not None
                else isinstance(settings[group][prop], str)
            )
        except ValueError as ex:
            msg = str(ex)
            passed = False

        if not passed:
            self.logger.warning(
                f"`{group}` invalid value for `{prop}` was found "
                f"in config.\n{msg}\nResorting to default value."
            )
            del settings[group][prop]
        return passed

    def is_config_present(self) -> bool:
        return self.config_path.is_file()

    def is_ssh_configured(self) -> bool:
        return self.ssh is not None

    def load_config(self):
        if self.is_config_present():
            with open(self.config_path, "r") as f:
                cfg = json.load(f)

            if "defaults" in cfg:
                defaults = cfg["defaults"]
                # import from json is a list by default, convert to tuple here
                if "cmdfiles" in defaults:
                    defaults["cmdfiles"] = tuple(defaults["cmdfiles"])
                # validate defaults
                self._validate_property(
                    settings=cfg, group="defaults", prop="username"
                )
                self._validate_property(
                    settings=cfg,
                    group="defaults",
                    prop="priority",
                    # ValueError raised if config string value not a member of JobPriority
                    test=lambda value: Config.JobPriority(value),
                )
                self._validate_property(
                    settings=cfg,
                    group="defaults",
                    prop="workers",
                    test=lambda value: isinstance(value, int) and value >= 0,
                    msg=f"must be an integer value >= 0.",
                )
                self._validate_property(
                    settings=cfg,
                    group="defaults",
                    prop="cmdfiles",
                    test=lambda value: isinstance(value, tuple)
                    and all([isinstance(e, str) for e in value]),
                    msg=f"must be a list or tuple of strings.",
                )
                self.defaults = Config.Defaults(**cfg["defaults"])

            if "ssh" in cfg and all(
                [
                    e in cfg["ssh"]
                    and self._validate_property(
                        settings=cfg, group="ssh", prop=e
                    )
                    for e in ["server", "username", "password"]
                ]
            ):
                ssh = cfg["ssh"]
                self.ssh = Config.SSHConfig(
                    server=ssh["server"],
                    username=ssh["username"],
                    password=ssh["password"],
                )
            else:
                self.ssh = None

            if (
                "plugins" in cfg
                and isinstance(cfg["plugins"], (list, tuple))
                and all([isinstance(e, str) for e in cfg["plugins"]])
            ):
                self.plugins = cfg["plugins"]
            else:
                self.plugins = []


def get_config() -> Config:
    global __config
    if __config is None:
        __config = Config()
    return __config



from typing import Dict, List

import pluggy

from solitude import TOOL_NAME

hookspec = pluggy.HookspecMarker(TOOL_NAME.lower())


@hookspec
def matches_command(cmd: str) -> bool:
    """Should this command be processed by this plugin?

    :param cmd: the command to test
    :return: True if command matches False otherwise
    """


@hookspec
def get_command_hash(cmd: str) -> str:
    """Computes the hash for the command
    This is used to uniquely link job status to commands.
    So if the exact same command is found they both link to the same job.
    Therefore it is recommended to remove parts from cmd that do not change
    the final results for the job.
    If you are uncertain what to do just return `cmd` as hash

    :param cmd: the command to compute the hash for
    :return: the command hash
    """


@hookspec
def retrieve_state(cmd: str) -> Dict:
    """Retrieve state for the job which can be set in a dictionary

    :param cmd: the command to test
    :return: a dictionary with the retrieved state (used in other calls)
    """


@hookspec
def is_command_job_done(cmd: str, state: Dict) -> bool:
    """Checks if the command has finished

    :param cmd: the command to test
    :param state: the retrieved state dictionary for this job
    :return: True if job is done False otherwise
    """


@hookspec
def get_command_status_str(cmd: str, state: Dict) -> str:
    """Retrieve state for the job which can be set in a dictionary

    :param cmd: the command to test
    :param state: the retrieved state dictionary for this job
    :return: a string containing job information and progress status
    """


@hookspec
def get_errors_from_log(log: str) -> List[str]:
    """Checks the log for errors

    :param log: the log string to parse
    :return: A list of error messages, empty list if no errors were found
    """



from typing import Any, Callable, Dict, List, Optional, Tuple
from types import ModuleType

import json
import sys
import tqdm
import re
import logging
import datetime
from pathlib import Path

from solitude import TOOL_NAME, hookspecs
from solitude.cache import Cache
from solitude.commands import CommandBase
import solitude.utils.ssh as ssh
import solitude.plugins.default
import solitude.plugins.csubmit
import solitude.plugins.interactive

from multiprocessing.pool import ThreadPool

import pluggy

from solitude.config import Config
from solitude.slurmjob import SlurmJob

logger = logging.getLogger(f"{TOOL_NAME}")


COMMENT_MARKER = "#"
USER_MACRO = "{user}"
SOL_DOCKER_REPO_MACRO = "{sol-docker-repo}"
# these are all aliases for the same repository
SOL_DOCKER_REPO_ALIASES = ["doduo1.umcn.nl", "doduo2.umcn.nl", "doduo.umcn.nl"]
# this is the repository that will be used at core.run
SOL_DOCKER_REPO = "doduo1.umcn.nl"


def resolve_cmd_files(cmd_files: List[Path]) -> List[str]:
    commands: List[str] = []
    for cmd_file in cmd_files:
        commands = commands + resolve_cmd_file(cmd_file=cmd_file)
    return commands


def resolve_cmd_file(cmd_file: Path) -> List[str]:
    try:
        with open(cmd_file, "r") as f:
            data = f.read()
        # auto-replaces hard-coded SOL_DOCKER_REPO_ALIASES to the SOL_DOCKER_REPO_MACRO
        # this makes started jobs still link to the macro under the hood
        # and facilitates transition to using the macro
        for repo_url in SOL_DOCKER_REPO_ALIASES:
            data = data.replace(repo_url, SOL_DOCKER_REPO_MACRO)
        # extract commands per line
        data_lines = [line.strip() for line in data.splitlines(keepends=False)]
        return [d for d in data_lines if any(d) and d[0] != COMMENT_MARKER]
    except FileNotFoundError:
        logger.error(
            f"ERROR: Could not find command file: {cmd_file}, skipping..."
        )
        return []


def shared_section(
    cmd_files: List[Path], cache_file: Path, cfg: Config
) -> Tuple[List[CommandBase], Cache, Optional[Config.SSHConfig]]:
    logger.debug(f"Starting action at: {datetime.datetime.now()}")
    # load plugins
    plugin_manager = setup_plugins(plugins=cfg.plugins)
    logger.debug(f"Read {len(plugin_manager.get_plugins())} command plugins")
    # load cache file
    cache: Cache = Cache(data_class=SlurmJob, file_name=cache_file)
    # perform cache auto cleanup (removes expired jobs)
    cache.cleanup()
    # load command files
    cmds_strs = resolve_cmd_files(cmd_files=cmd_files)
    cmds: List[CommandBase] = read_commands(
        commands=cmds_strs, cache=cache, plugin_manager=plugin_manager,
    )
    # retrieve ssh credentials
    return cmds, cache, cfg.ssh


class SpecialFormatter(logging.Formatter):
    def __init__(self, *args, **kwargs):
        super(SpecialFormatter, self).__init__(*args, **kwargs)
        self._main_formatter = logging.Formatter(fmt="%(message)s")
        self._ssh_formatter = logging.Formatter(fmt="[SSHClient] %(message)s")
        self._special_formatter = logging.Formatter(
            fmt="%(asctime)s [%(name)-10s] %(message)s",
            datefmt="%y-%m-%d %H:%M:%S",
        )

    def format(self, record: logging.LogRecord) -> str:
        if record.name == TOOL_NAME:
            return self._main_formatter.format(record=record)
        elif record.name == "SSHClient":
            return self._ssh_formatter.format(record=record)
        else:
            return self._special_formatter.format(record=record)


def log_setup(log_level: int = logging.DEBUG) -> logging.Logger:
    """setup some basic logging"""
    log = logging.getLogger("")
    log.setLevel(log_level)
    fmt = SpecialFormatter()
    ch = logging.StreamHandler(sys.stdout)
    ch.setLevel(log_level)
    ch.setFormatter(fmt)
    log.addHandler(ch)
    logging.getLogger("paramiko.transport").setLevel(logging.ERROR)
    logging.getLogger("urllib3.connectionpool").setLevel(logging.ERROR)
    return logging.getLogger(TOOL_NAME)


def get_attr_from_module(pclass: str) -> Any:
    splitted: List[str] = pclass.rsplit(".", 1)
    mod = __import__(splitted[0], fromlist=[str(splitted[1])])
    return getattr(mod, splitted[1])


def parse_plugins(plugins: List[str]) -> List[ModuleType]:
    results = []
    for plugin_name in plugins:
        if len(plugin_name.strip()) > 0:
            try:
                results.append(get_attr_from_module(plugin_name.strip()))
            except ModuleNotFoundError:
                logger.error(
                    f"!WARNING! Could not find plugin listed in config: {plugin_name}"
                )
    return results


def match_plugin(pm: pluggy.PluginManager, cmd: str) -> ModuleType:
    try:
        idx = pm.hook.matches_command(cmd=cmd).index(True)
    except ValueError:
        raise ValueError(
            f"No plugin found for handling this type of command: {cmd}"
        )
    return pm.hook.matches_command.get_hookimpls()[::-1][idx].plugin


def read_commands(
    commands: List[str], cache: Cache, plugin_manager: pluggy.PluginManager
) -> List[CommandBase]:
    return [
        CommandBase(
            cmd=x, plugin=match_plugin(pm=plugin_manager, cmd=x), cache=cache
        )
        for x in commands
    ]


def execute_commands(
    commands: List[CommandBase],
    map_fn: Callable,
    workers: int,
    show_progress_bar: bool = True,
):
    if workers > 0:
        tpex = ThreadPool(processes=workers)
        list(
            tqdm.tqdm(
                tpex.imap(map_fn, commands),
                total=len(commands),
                desc="Processing cmds ({} threads)".format(workers),
                disable=not show_progress_bar,
            )
        )
    else:
        for cmd in tqdm.tqdm(
            commands,
            desc="Processing cmds (no threads)",
            disable=not show_progress_bar,
        ):
            map_fn(cmd)
    if show_progress_bar:
        logger.info("")


def read_cache(file_name: Path) -> Dict:
    if not file_name.resolve().is_file():
        return {}
    with open(file_name, "r") as f:
        data = json.load(f)
    return data


def extend(
    cmd_files: List[Path],
    jobids: List[int],
    workers: int,
    cfg: Config,
    cache_file: Path,
):
    _run_command_on_running_job(
        cmd_files,
        jobids,
        workers,
        cfg,
        cache_file,
        command="c-extend",
        action_str="EXTENDING",
    )


def stop(
    cmd_files: List[Path],
    jobids: List[int],
    workers: int,
    cfg: Config,
    cache_file: Path,
):
    _run_command_on_running_job(
        cmd_files,
        jobids,
        workers,
        cfg,
        cache_file,
        command="c-stop",
        action_str="STOPPING",
    )


def link(
    cmd_files: List[Path],
    jobid: int,
    slurm_jobid: int,
    cfg: Config,
    cache_file: Path,
) -> bool:
    commands, cache, _ = shared_section(
        cmd_files=cmd_files, cache_file=cache_file, cfg=cfg
    )
    cmd = commands[jobid - 1]
    # link jobid to slurm_jobid
    if SlurmJob.check_if_job_exists(jobid=slurm_jobid):
        job = SlurmJob(jobid=slurm_jobid)
        job.update()
        cache.update(key=cmd.hash, value=job)
        return True
    return False


def unlink(cmd_files: List[Path], jobid: int, cfg, cache_file: Path) -> bool:
    commands, cache, _ = shared_section(
        cmd_files=cmd_files, cache_file=cache_file, cfg=cfg
    )
    cmd = commands[jobid - 1]
    if cmd.hash in cache:
        cache.delete(key=cmd.hash)
        return True
    return False


def _run_command_on_running_job(
    cmd_files: List[Path],
    jobids: List[int],
    workers: int,
    cfg: Config,
    cache_file: Path,
    command: str,
    action_str: str,
):
    ssh_client = ssh.SSHClient()
    assert cfg.ssh is not None
    ssh_client.connect(*cfg.ssh)

    commands, _, credentials = shared_section(
        cmd_files=cmd_files, cache_file=cache_file, cfg=cfg
    )

    # update selected commands only (use workers)
    subcmds = [cmd for idx, cmd in enumerate(commands) if idx + 1 in jobids]

    def _update_cmd(cmd):
        cmd.update_job_info()

    execute_commands(
        subcmds, map_fn=_update_cmd, workers=workers, show_progress_bar=True
    )

    for idx, cmd in zip(jobids, subcmds):
        if cmd.is_running():
            assert cmd.job_info is not None
            logger.info(f"ATTEMPT {action_str} Job: {idx}")
            ssh_client.exec_command(f"./{command} {cmd.job_info.id}")
        else:
            logger.info(f"Job {idx} is no longer running. Skipping...")
    log_action_finished()


def list_jobs(
    cmd_files: List[Path],
    jobids: List[int],
    workers: int,
    cfg: Config,
    cache_file: Path,
    selected_only: bool = False,
    show_progress_bar: bool = True,
):
    commands, _, _ = shared_section(
        cmd_files=cmd_files, cache_file=cache_file, cfg=cfg
    )

    # update selected commands only (use workers)
    selected_ids = jobids
    if not selected_only:
        jobids = [e for e in range(1, len(commands) + 1)]
    subcmds = [
        cmd
        for idx, cmd in enumerate(commands)
        if (idx + 1 in jobids) or not selected_only
    ]

    def _update_cmd(cmd):
        cmd.update()

    execute_commands(
        subcmds,
        map_fn=_update_cmd,
        workers=workers,
        show_progress_bar=show_progress_bar,
    )
    logger.info("======Commands======")
    logger.info(CommandBase.header_str())
    for idx, cmd in zip(jobids, subcmds):
        logger.info(f'{idx:2d}{"*" if idx in selected_ids else " "} {cmd}')
    log_action_finished()


def log_action_finished():
    logger.debug(f"Finished action at: {datetime.datetime.now()}")


def run(
    cmd_files: List[Path],
    jobids: List[int],
    workers: int,
    user: str,
    priority: str,
    reservation: str,
    duration: int,
    ignore_errors: bool,
    cfg: Config,
    cache_file: Path,
):
    ssh_client = ssh.SSHClient()
    assert cfg.ssh is not None
    ssh_client.connect(*cfg.ssh)

    commands, cache, credentials = shared_section(
        cmd_files=cmd_files, cache_file=cache_file, cfg=cfg
    )

    def _update_cmd(cmd):
        cmd.update_job_info()

    execute_commands(
        commands, map_fn=_update_cmd, workers=workers, show_progress_bar=False
    )

    high_priority_job_available = not any(
        [
            c.job_info is not None
            and c.is_running()
            and c.job_info.priority == "high"
            for c in commands
            if c.job_info is not None and c.job_info.user == user
        ]
    )
    if high_priority_job_available:
        logger.info(f"No high priority job running for user {user}!")
    elif priority == "high":
        logger.info(
            f"No high priority job available for user {user}! Switching to low priority..."
        )
        priority = "low"

    # update selected commands only (use workers)
    subcmds = [cmd for idx, cmd in enumerate(commands) if idx + 1 in jobids]

    def _update_cmd2(cmd):
        cmd.update_state()
        cmd.update_errors()

    execute_commands(
        subcmds, map_fn=_update_cmd2, workers=workers, show_progress_bar=True
    )

    for idx, cmd in zip(jobids, subcmds):
        logger.info(f"ATTEMPT EXECUTE Job {idx}")
        if cmd.is_running():
            logger.info("Job is already running. skipping...")
            logger.info(str(cmd))
        elif cmd.is_finished():
            logger.info("Job has already finished. skipping...")
            logger.info(str(cmd))
        elif cmd.is_erroneous() and not ignore_errors:
            logger.info(
                "Job has halted with errors. Use --ignore-errors to run anyway. skipping..."
            )
            logger.info(str(cmd))
        else:
            try:
                # set correct priority level in command
                cmdstr = cmd.cmd
                if "c-submit " in cmdstr:
                    # remove any priority optional flags
                    cmdstr = re.sub(r"--priority=[a-zA-Z0-9]*\s", "", cmdstr)
                    # add desired priority flag after c-submit command
                    cmdstr = re.sub(
                        r"c-submit",
                        "c-submit --priority={}".format(priority),
                        cmdstr,
                    )
                    if duration is not None:
                        cmdstr = re.sub(
                            r"\s(\d+)\soni:11500",
                            " {} oni:11500".format(str(duration)),
                            cmdstr,
                        )
                    if reservation is not None:
                        cmdstr = re.sub(
                            r"c-submit",
                            "c-submit --reservation={}".format(reservation),
                            cmdstr,
                        )
                cmdstr = cmdstr.replace(USER_MACRO, user)
                cmdstr = cmdstr.replace(SOL_DOCKER_REPO_MACRO, SOL_DOCKER_REPO)

                # attempt schedule command
                result = ssh_client.exec_command(cmdstr)
                if result is not None and ("c-submit " in cmdstr):
                    link_id = int(result[0].strip().split(" ")[-1])
                    # can have only one high priority job so switch to low priority afterwards
                    cache.update(
                        key=cmd.hash,
                        value=SlurmJob(
                            jobid=link_id, user=user, priority=priority
                        ),
                    )
                    if priority == "high":
                        logger.info(
                            "Assigned a high priority job switching to low priority now"
                        )
                        priority = "low"
                    logger.info("http://oni:11080/show/{}".format(link_id))
            except Exception as e:
                raise e

    log_action_finished()


def setup_plugins(
    plugins: List[str], debug: bool = False
) -> pluggy.PluginManager:
    plugin_manager = pluggy.PluginManager(TOOL_NAME)
    plugin_manager.add_hookspecs(hookspecs)
    # Add default fallback plugins
    plugin_manager.register(solitude.plugins.default)
    plugin_manager.register(solitude.plugins.csubmit)
    plugin_manager.register(solitude.plugins.interactive)
    # Load externally packaged plugins
    plugin_manager.load_setuptools_entrypoints(group=TOOL_NAME)
    # Add additional plugins from cfg
    plugins_mods: List[ModuleType] = parse_plugins(plugins=plugins)
    for plugin in reversed(plugins_mods):
        plugin_manager.register(plugin)
    # validate plugins (must strictly implement all plugin methods)
    plugin_manager.check_pending()
    for plugin in plugin_manager.get_plugins():
        validate_plugin(plugin)
    if debug:
        plugin_manager.trace.root.setwriter(print)
        plugin_manager.enable_tracing()
    return plugin_manager


def validate_plugin(plugin: ModuleType):
    required_fns = [
        e for e in dir(solitude.plugins.default) if "__" not in e and "_" in e
    ]
    for fn in required_fns:
        if not hasattr(plugin, fn):
            raise NotImplementedError(
                f"Plugin {plugin} has no hook_implementation for {fn}"
            )



import re
from typing import Dict, List
from solitude import hookimpl


@hookimpl
def matches_command(cmd: str) -> bool:
    return True


@hookimpl
def get_command_hash(cmd: str) -> str:
    return cmd


@hookimpl
def retrieve_state(cmd: str) -> Dict:
    return {}


@hookimpl
def is_command_job_done(cmd: str, state: Dict) -> bool:
    return False


@hookimpl
def get_command_status_str(cmd: str, state: Dict) -> str:
    return cmd


@hookimpl
def get_errors_from_log(log: str) -> List[str]:
    errors = []
    if log.find("Killed") != -1:
        errors.append("NoResource")
    if log.find("CUDA_ERROR") != -1:
        errors.append("CUDA")
    errors = errors + [e[1:-1] for e in re.findall(r"\s.*Error:", log)]
    return errors



import socket
import sys
import logging

import paramiko


class SSHClient(object):
    def __init__(self, exit_on_fail=True):
        self.logger = logging.getLogger(type(self).__name__)
        self._exit_on_fail = exit_on_fail
        self._ssh = paramiko.SSHClient()
        self._ssh.load_system_host_keys()
        self._ssh.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        self._connected = False

    def connect(self, server, username, password) -> bool:
        try:
            self._ssh.connect(
                hostname=server, username=username, password=password
            )
            self._connected = True
            return True
        except socket.gaierror:
            self.logger.error(
                "Failed to connect to server: {}@{} (socket.gaierror)".format(
                    username, server
                )
            )
            if self._exit_on_fail:
                self.logger.error(
                    "Run `solitude config test` to test the current configuration."
                )
                sys.exit()
        except Exception as e:
            if self._exit_on_fail:
                raise e
            else:
                self.logger.error("Failed with: {}".format(e))
        return False

    def is_connected(self) -> bool:
        return self._connected

    def exec_command(self, cmd_to_execute):
        if not self.is_connected():
            if not self.connect():
                return None
        self.logger.info("CMD: {}".format(cmd_to_execute))
        try:
            ssh_stdin, ssh_stdout, ssh_stderr = self._ssh.exec_command(
                cmd_to_execute
            )
            result = (
                "".join(ssh_stdout.readlines()),
                "".join(ssh_stderr.readlines()),
            )
            self.logger.info("RESULT: {}".format(result[0]))
            return result
        except Exception as e:
            self.logger.error("Failed with: {}".format(e))
            if self._exit_on_fail:
                sys.exit()
        return None



import json
import logging
import tempfile
import time
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, Optional, Type


class CacheData(ABC):
    DEFAULT_EXPIRATION_TIME = 1209600  # 2 weeks expressed in secs

    def __init__(self, timestamp: Optional[int] = None):
        self.timestamp = int(
            timestamp if timestamp is not None else time.time()
        )

    @abstractmethod
    def to_dict(self) -> Dict:
        pass

    @classmethod
    @abstractmethod
    def from_dict(cls: Type, dic: Dict) -> "CacheData":
        pass


class Cache(object):
    def __init__(self, data_class: Type[CacheData], file_name: Path):
        self._logger = logging.getLogger("Cache")
        self.__cache: Dict[str, CacheData] = {}
        self._file_name = file_name
        self.__data_class = data_class
        self._load()

    def _load(self):
        self.__cache.clear()
        if not self._file_name.resolve().is_file():
            self._logger.debug(f"{self._file_name} does not exists.")
            return
        try:
            with open(self._file_name, "r") as f:
                data = json.load(f)
        except json.JSONDecodeError as ex:
            self._logger.error(
                f"Error reading the Cache file (JSONDecodeError) : {self._file_name}"
            )
            self._logger.debug(f"{ex}")
            return
        for key, value in data.items():
            self.__cache[key] = self.__data_class.from_dict(dic=value)
        self._logger.debug(
            f"Read {len(self)} cached {self.__data_class} items"
        )

    def update(self, key: str, value: CacheData):
        self.__cache[key] = value
        self._save()

    def delete(self, key: str):
        if key in self.__cache:
            del self.__cache[key]
            self._save()

    def __getitem__(self, key: str) -> CacheData:
        return self.__cache[key]

    def __len__(self) -> int:
        return len(self.__cache)

    def __contains__(self, key: str) -> bool:
        return key in self.__cache

    def _save(self):
        self._file_name.parent.mkdir(parents=True, exist_ok=True)
        data = {k: v.to_dict() for k, v in self.__cache.items()}
        self._logger.debug(
            f"Writing {self.__data_class} items to cache: {self._file_name}"
        )
        json_write_atomic(fname=self._file_name, data=data)

    def cleanup(
        self, expiration_time: int = CacheData.DEFAULT_EXPIRATION_TIME
    ):
        current_time = time.time()
        initial_size = len(self.__cache)
        for key in list(self.__cache.keys()):
            if current_time - self.__cache[key].timestamp >= expiration_time:
                del self.__cache[key]
        cleaned_items = initial_size - len(self.__cache)
        if cleaned_items > 0:
            self._logger.debug(
                f"Removed {cleaned_items} expired {self.__data_class} items from cache."
            )
            self._save()


def json_write_atomic(fname: Path, data: Any):
    # specify dir to ensure same filesystem...
    with tempfile.NamedTemporaryFile(
        mode="w", delete=False, dir=str(fname.parent)
    ) as f:
        tmp_filename = Path(f.name)
        json.dump(data, f, indent=4)
    # atomic write with replacement (most of the time)
    tmp_filename.replace(fname)



from os import path
from setuptools import setup, find_packages

VERSION = "0.2.0"

this_directory = path.abspath(path.dirname(__file__))

with open(
    path.join(this_directory, "README.md"), mode="r", encoding="utf-8"
) as fh:
    long_description = fh.read()

with open(
    path.join(this_directory, "requirements.txt"), mode="r", encoding="utf-8"
) as fh:
    requirements = [e.strip() for e in fh.readlines() if e.strip() != ""]

setup(
    name="solitude",
    version=VERSION,
    author="S.C. van de Leemput",
    author_email="sil.vandeleemput@radboudumc.nl",
    install_requires=requirements,
    license="LICENSE",
    entry_points={"console_scripts": ["solitude=solitude.cli:main"]},
    packages=find_packages(),
    description="A simple light-weight extendable command line tool for managing jobs on DIAG's SOL cluster.",
    long_description_content_type="text/markdown",
    long_description=long_description,
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
    ],
)
